(in-package #:cl-yahoo)

;;;;; MACROS

(defmacro define-result-container (c-name &rest slots);; defclass-wrapper,
  "Expand into DEFCLASS and make initargs and accessors for slots.
   Ex. 
   (define-result-container c-name slot1 slot2) =>
   (defclass c-name ()
     ((slot1 :accessor slot1 :initarg :slot1)
      (slot2 :accessor slot2 :initarg :slot2)))."
  `(defclass ,c-name ()
     (,@(loop for slot in slots collect;; variable capture?
	     (list slot :accessor slot :initarg (add-colon slot))))))


(defmacro define-query-url ((&key type)
			    &rest keys)
  "Make a method that dispatches on TYPE, returning
   a list of two strings: the url and the queries."
  (with-gensyms (search-type rest)
    `(defmethod make-query-url ((,search-type (eql ,type)) &rest ,rest  &key ,@keys)
       (declare (ignorable ,@(cons search-type keys)))
       (values
	(get-search-url ,type)
	(make-search-string  ,rest)))))


(defmacro define-search-result ((&key type)
				&rest keys)
  "Make a function MAKE-type-RESULT that makes an instance of a result TYPE."
  `(defun ,(symcat 'make- type '-result) (result)
     (let ((result (make-assoc-list result)))
       (let (,@(loop for key in keys
		  for var in (mapcar #'lispify-var keys)
		  collect 
		  `(,var (second (assoc ,key result)))))
	 (make-instance (intern (string (symcat ',type '-result)) :cl-yahoo)
			,@(loop for sym in (mapcar #'lispify-var keys)
			     appending
			     (list (intern (string sym) :keyword) sym)))))))


(defmacro define-search-method ((&key type http-post) &rest keys)
  "Define a method YSEARCH that dispatches on TYPE, and makes a search on yahoo. If HTTP-POST is true, do a HTTP-POST."
  (with-gensyms (xml-result search-type url query)
    `(defmethod ysearch ((,search-type (eql ,type)) &key ,@keys)
       (declare (ignorable ,search-type))
       (multiple-value-bind (,url ,query)
	   (make-query-url
	    ,type
	    ,@(loop for key in keys append
		   (list (intern (string key):keyword) key)))
	 (let ((,xml-result
		(fetch-and-parse-xml ,url ,query 
				     ,@(when http-post (list :http-post t)))))
	   (when (assoc :|Error| ,xml-result)
	     (error 'yahoo-error :error-message 
		    (second (assoc :|Message| (remove-if-not #'consp ,xml-result)))))
	   (loop for result in (cdr ,xml-result)
	      for obj = (,(symcat 'make- type '-result) result)
	      when (or (slot-val-not-nil-p obj 'url)
		       (slot-val-not-nil-p obj 'result))
	      collect obj))))))

(defmacro define-ysearch (&key type request-parameters result-fields
			  http-post 
			  (print-method 'url))
  "Define the necessary ingredients for a Yahoo-search:
   1. A class and accessors for the result
   2. A method to parse the result into a result object
   3. A query-string 
   4. A search method
   5. A print method for the result object"
  `(progn 
     (define-result-container ,(symcat type '-result)
	 ,@(loop for field in result-fields collect
		(lispify-var field)))
     (define-search-result (:type ,type) ,@result-fields)
     (define-query-url (:type ,type)
	 ,@request-parameters)
     (define-search-method 
	 (:type ,type ,@(when http-post (list :http-post t)))
	 ,@request-parameters)
     (defmethod print-object ((result ,(symcat type '-result)) stream)
       (print-unreadable-object (result stream)
	 (format stream "~a" (,print-method result))))))
